# -*- coding: utf-8 -*-
from .l2 import L2
from .l2 import L2 as Line
from .l3 import L3
from .l3 import L3 as QuadraticLine
from .t3 import T3
from .t3 import T3 as Tri
from .q4 import Q4
from .q4 import Q4 as Quad
from .q8 import Q8
from .q9 import Q9
from .t6 import T6
from .h8 import H8
from .h8 import H8 as Hex
from .h27 import H27
from .tet4 import TET4
from .tet4 import TET4 as Tetra
from .tet10 import TET10
from .w6 import W6
from .w6 import W6 as Wedge
from .w18 import W18

__all__ = [
    "L2",
    "Line",
    "L3",
    "QuadraticLine",
    "T3",
    "Tri",
    "Q4",
    "Quad",
    "Q8",
    "Q9",
    "T6",
    "H8",
    "Hex",
    "H27",
    "TET4",
    "Tetra",
    "TET10",
    "W6",
    "Wedge",
    "W18",
]
